/*
 * Simple block device driver
 *   built on Fedora16 alpha(Linux 3.0)
 * 
 * [Install]
 *   # insmod ./bdev.ko
 *   # cat /proc/devices | grep bdev
 *   (# mknod /dev/bdev b 252 0)
 *   # mknod /dev/bdev1 b 252 1
 *   # fdisk /dev/bdev
 *
 * (2011.8.27 Yutaka Hirata)
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>

MODULE_LICENSE("GPL");

#define MODULE_NAME	"bdev"  /* driver name */
#define SECTOR_SIZE	512     /* sector size in bytes */
#define N_SECTORS	10240    /* total sector number */

static int bdev_major = 0;   /* 0=dynamic allocation major number */
static int bdev_npart = 0;   /* disk partition count */
module_param(bdev_npart, int, 0);

static struct request_queue *bdev_queue;

typedef struct bdev_device {
	unsigned long size;
	unsigned char *data;
	struct gendisk *gd;
} bdev_device_t;

static bdev_device_t bdev_device;


static int bdev_ioctl(struct block_device *bdev, fmode_t mode,
		unsigned int cmd, unsigned long arg)
{
	struct hd_geometry geo;
	int ret = 0;

	printk("%s entered\n", __func__);

	switch (cmd) {
		case HDIO_GETGEO:
			memset(&geo, 0, sizeof(geo));
			geo.heads = 4;
			geo.sectors = 16;
			geo.start = 4;
			geo.cylinders = (bdev_device.size & ~0x3F) / (4 * 16);
			printk("%lu %u\n", bdev_device.size & ~0x3F, geo.cylinders);
			if (copy_to_user((void *)arg, &geo, sizeof(geo))) {
				ret = -EFAULT;
			}
			break;

		default:
			ret = -ENOTTY;
			break;
	}

	return (ret);
}

static struct block_device_operations bdev_fops = {
	.owner = THIS_MODULE,
	.ioctl = bdev_ioctl,
};

static int do_transfer_bdev(int rw, struct bio_vec *vec, sector_t sector)
{
	unsigned int vec_offset = vec->bv_offset;
	int size = vec->bv_len;
	int offset = sector * 512;
	char *src, *dst;

	if (rw == READ) {
		src = bdev_device.data + offset;
		dst = kmap(vec->bv_page) + vec_offset;

	} else {
		dst = bdev_device.data + offset;
		src = kmap(vec->bv_page) + vec_offset;

	}

	memcpy(dst, src, size);
	kunmap(vec->bv_page);

	if (rw == READ) {
		flush_dcache_page(vec->bv_page);
	}

	return 0;
}

static int do_make_request(struct request_queue *q, struct bio *bio)
{
	struct block_device *bdev = bio->bi_bdev;
	sector_t sector = bio->bi_sector;
	unsigned long len = bio->bi_size / SECTOR_SIZE;
	struct bio_vec *bvec;
	int rw, i, ret;

	rw = bio_data_dir(bio);

#if 0
	printk("%s: sector %llu size %u len %lu rw %d\n", 
			__func__, sector, bio->bi_size, len, rw);
#endif

	if (sector + len > get_capacity(bdev->bd_disk)) {
		goto error;
	}

	ret = 0;
	bio_for_each_segment(bvec, bio, i) {
		ret |= do_transfer_bdev(rw, bvec, sector);
		sector += bvec->bv_len / 512;
	}
	if (ret)
		goto error;

	bio_endio(bio, 0);
	return 0;

error:
	bio_io_error(bio);

	return 0;
}

static void free_driver_resources(void)
{
	/* remove disk */
	if (bdev_device.gd) {
		del_gendisk(bdev_device.gd);
		put_disk(bdev_device.gd);
	}

	/* remove queue */
	if (bdev_queue) {
		blk_cleanup_queue(bdev_queue);
	}

	if (bdev_major > 0)
		unregister_blkdev(bdev_major, MODULE_NAME);

	if (bdev_device.data) 
		//kfree(bdev_device.data);
		vfree(bdev_device.data);
}

static int bdev_init(void)
{
	struct gendisk *gd = NULL;
	int ret = 0;
	int major;

	bdev_device.size = SECTOR_SIZE * N_SECTORS;
	//bdev_device.data = kzalloc(bdev_device.size, GFP_KERNEL);
	bdev_device.data = vmalloc(bdev_device.size);
	if (bdev_device.data == NULL) {
		printk("error1\n");
		ret = -ENOMEM;
		goto error;
	}

	bdev_queue = blk_alloc_queue(GFP_KERNEL);
	if (bdev_queue == NULL) {
		printk("error2\n");
		ret = -ENOMEM;
		goto error;
	}
	blk_queue_make_request(bdev_queue, do_make_request);

	major = register_blkdev(bdev_major, MODULE_NAME);
	if (major < 0) {
		printk("error3\n");
		ret = major;
		goto error;
	}
	if (bdev_major == 0)
		bdev_major = major;

	bdev_device.gd = gd = alloc_disk(bdev_npart + 1);
	if (gd == NULL) {
		printk("error4\n");
		ret = -ENOMEM;
		goto error;
	}
	gd->major = bdev_major;
	gd->first_minor = 0;
	gd->fops = &bdev_fops;
	gd->private_data = &bdev_device;
	gd->queue = bdev_queue;
	set_capacity(gd, N_SECTORS);
	strlcpy(gd->disk_name, "bdev", sizeof(gd->disk_name));

	add_disk(gd);

	printk("%s: bdev driver loaded\n", __func__);

	return (0);

error:
	free_driver_resources();

	return (ret);
}

static void bdev_exit(void)
{
	free_driver_resources();

	printk("%s: bdev driver Unloaded\n", __func__);
}

module_init(bdev_init);
module_exit(bdev_exit);

